import pytest
import awscost
from awscost import commands
from click.testing import CliRunner


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


def test_show_version(runner):
    result = runner.invoke(commands.cli, ["-v"])
    assert result.exit_code == 0
    assert result.output.strip() == awscost.VERSION



import yaml
import os
from tabulate import tabulate
from datetime import datetime
from collections import OrderedDict
from .logger import get_logger
from .cost_explorer_client import CostExplorerClient
from . import constants
from .date_util import DateUtil


class CostExplorer:
    """
    convert responce data class.
    """

    def __init__(
        self,
        config=None,
        profile=None,
        granularity=None,
        point=None,
        start=None,
        end=None,
        dimensions=None,
        tags=None,
        filter=None,
        metrics=None,
        aws_profile=None,
        debug=None,
        total=None,
        threshold=None,
    ):
        # read profile
        profile = self._read_profile(config, profile)

        self.granularity = self.get_not_none_first(
            granularity, profile.get("granularity"), constants.DEFAULT_GRANULARITY
        )
        self.dimensions = self.get_not_none_first(
            dimensions, profile.get("dimensions"), constants.DEFAULT_DIMENSIONS
        )
        self.tags = self.get_not_none_first(tags, profile.get("tags"), [])
        self.metrics = self.get_not_none_first(
            metrics, profile.get("metrics"), constants.DEFAULT_METRICS
        )
        self.total = self.get_not_none_first(
            total, profile.get("total"), constants.DEFAULT_TOTAL
        )
        debug = self.get_not_none_first(
            debug, profile.get("debug"), constants.DEFAULT_DEBUG
        )
        self.logger = get_logger(debug=debug)
        self.threshold = self.get_not_none_first(
            threshold, profile.get("threshold"), constants.DEFAULT_THRESHOLD
        )

        aws_profile = self.get_not_none_first(aws_profile, profile.get("aws_profile"))
        filter = self.get_not_none_first(filter, profile.get("filter"))
        point = self.get_not_none_first(
            point, profile.get("point"), constants.DEFAULT_POINT
        )
        start = self.get_not_none_first(
            start, profile.get("start"), DateUtil.get_start(self.granularity, point)
        )
        end = self.get_not_none_first(
            end, profile.get("end"), datetime.today().strftime("%Y-%m-%d")
        )

        self.cost_explorer_client = CostExplorerClient(
            self.granularity,
            start,
            end,
            filter=filter,
            metrics=metrics,
            aws_profile=aws_profile,
            debug=debug,
        )

    def to_tabulate(self, tablefmt=None):
        """
        convert tabulate style.
        """
        data = self.get_cost_and_usage_total_and_group_by()
        converts = []
        for k, amounts in data.items():
            converts.append(dict({"key": k}, **amounts))
        last_time = list(converts[0].keys())[-1]
        converts = sorted(
            converts,
            key=lambda x: 0 if x.get(last_time) is None else x.get(last_time),
            reverse=True,
        )
        return tabulate(converts, headers="keys", tablefmt=tablefmt)

    def get_cost_and_usage_total_and_group_by(self):
        """
        Get cost time series data by specifying start and end, and merge total and group_by

        """
        # totalを取得
        total = self.get_cost_and_usage_total()

        # group byしたデータを取得
        group_by_results = self.get_cost_and_usage_group_by()

        # totalと0埋めしたgroup byをmergeする
        group_by_results_pad_zero = self.__class__.pad_zero(total, group_by_results)
        if self.total:
            merged = OrderedDict(total, **group_by_results_pad_zero)
            return merged
        return group_by_results_pad_zero

    def get_cost_and_usage_total(self):
        """
        Get cost time series data by specifying start and end
        """
        # totalを取得
        cost_and_usage = self.cost_explorer_client.get_cost_and_usage()
        total = self._convert_results_total(cost_and_usage)
        return total

    def get_cost_and_usage_group_by(self):
        """
        Get cost time series data by specifying start and end
        """
        cost_and_usage_per_service = self.cost_explorer_client.get_cost_and_usage(
            dimensions=self.dimensions, tags=self.tags
        )
        results = self._convert_results_group_by(cost_and_usage_per_service)
        results = dict(
            filter(
                lambda item: max(list(item[1].values())) >= self.threshold,
                results.items(),
            )
        )
        return results

    def _convert_results_group_by(self, cost_and_usage_per_service):
        """
        Parse data structure for which group-by is specified
        """
        results = OrderedDict()
        for result in cost_and_usage_per_service:
            start_period = result.get("TimePeriod").get("Start")
            time_key = self._convert_period(start_period)
            groups = result.get("Groups")
            for group in groups:
                group_by_key = ",".join(group.get("Keys"))
                if results.get(group_by_key) is None:
                    results[group_by_key] = OrderedDict()
                else:
                    results[group_by_key] = results.get(group_by_key)
                metrics = group.get("Metrics")
                amount = metrics.get(self.metrics).get("Amount")
                results[group_by_key][time_key] = round(float(amount), 2)
        return results

    def _convert_results_total(self, cost_and_usage_per_service):
        """
        Parse the data structure of Total
        """
        results = OrderedDict([("Total", OrderedDict())])
        for result in cost_and_usage_per_service:
            start_period = result.get("TimePeriod").get("Start")
            time_key = self._convert_period(start_period)
            metrics = result.get("Total")
            amount = metrics.get(self.metrics).get("Amount")
            results["Total"][time_key] = round(float(amount), 2)
        return results

    def _convert_period(self, start_period):
        """
        Switching the date displayed in the header between monthly and daily
        """
        if self.granularity == "MONTHLY":
            return datetime.strptime(start_period, "%Y-%m-%d").strftime("%Y-%m")
        return datetime.strptime(start_period, "%Y-%m-%d").strftime("%m-%d")

    def _read_profile(self, config, profile_name):
        config = config or constants.DEFAULT_CONFIG
        profile_name = profile_name or constants.DEFAULT_PROFILE
        if config and os.path.exists(config):
            profile = yaml.safe_load(open(config, encoding="UTF-8").read()).get(
                profile_name
            )
            if profile is None:
                profile = {}
        else:
            profile = {}
        return profile

    @staticmethod
    def pad_zero(total, group_by_results):
        """
        Fill the value to zero
        """
        # 0埋めするためのdictを作成
        pad_zero = OrderedDict()
        for k, v in total.get("Total").items():
            pad_zero[k] = 0

        group_by_results_pad_zero = OrderedDict()
        for k, v in group_by_results.items():
            merged = OrderedDict(pad_zero, **v)
            group_by_results_pad_zero[k] = merged
        return group_by_results_pad_zero

    def get_not_none_first(self, *args):
        for v in list(args):
            if v is not None:
                return v
        return None



from awscost.cost_explorer import CostExplorer
from awscost.cost_explorer_client import CostExplorerClient
from mock import patch
from collections import OrderedDict


class TestCostExplorer(object):
    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    def test_pad_zero(self):
        total = {"Total": {"2020-01": 1.5, "2020-02": 1.5, "2020-03": 1.5}}
        group_by_results = {
            "EC2 - Other": {"2020-02": 1.5},
            "Amazon Simple Storage Service": {"2020-01": 2.5},
        }
        group_by_results_pad_zero = CostExplorer.pad_zero(total, group_by_results)

        ec2 = group_by_results_pad_zero.get("EC2 - Other")
        assert ec2.get("2020-01") == 0
        assert ec2.get("2020-02") == 1.5
        assert ec2.get("2020-03") == 0

        s3 = group_by_results_pad_zero.get("Amazon Simple Storage Service")
        assert s3.get("2020-01") == 2.5
        assert s3.get("2020-02") == 0
        assert s3.get("2020-03") == 0

    def test_get_cost_and_usage_group_by(self):
        cost_explorer = CostExplorer()
        with patch.object(CostExplorerClient, "get_cost_and_usage") as mock_foo:
            mock_foo.return_value = [
                {
                    "TimePeriod": {"Start": "2019-12-01", "End": "2020-01-01"},
                    "Total": {},
                    "Groups": [
                        {
                            "Keys": ["AWS CloudTrail"],
                            "Metrics": {
                                "UnblendedCost": {"Amount": "4.380886", "Unit": "USD"}
                            },
                        },
                    ],
                    "Estimated": False,
                },
                {
                    "TimePeriod": {"Start": "2020-01-01", "End": "2020-02-01"},
                    "Total": {},
                    "Groups": [
                        {
                            "Keys": ["AWS CloudTrail"],
                            "Metrics": {
                                "UnblendedCost": {"Amount": "4.380886", "Unit": "USD"}
                            },
                        },
                        {
                            "Keys": ["AWS Cost Explorer"],
                            "Metrics": {
                                "UnblendedCost": {"Amount": "0.02", "Unit": "USD"}
                            },
                        },
                        {
                            "Keys": ["AWS Key Management Service"],
                            "Metrics": {
                                "UnblendedCost": {
                                    "Amount": "4.002411936",
                                    "Unit": "USD",
                                }
                            },
                        },
                        {
                            "Keys": ["AWS Lambda"],
                            "Metrics": {
                                "UnblendedCost": {
                                    "Amount": "0.0010751459",
                                    "Unit": "USD",
                                }
                            },
                        },
                    ],
                    "Estimated": False,
                },
            ]
            assert cost_explorer.get_cost_and_usage_group_by() == OrderedDict(
                [
                    (
                        "AWS CloudTrail",
                        OrderedDict([("2019-12", 4.38), ("2020-01", 4.38)]),
                    ),
                    ("AWS Cost Explorer", OrderedDict([("2020-01", 0.02)])),
                    ("AWS Key Management Service", OrderedDict([("2020-01", 4.0)])),
                    ("AWS Lambda", OrderedDict([("2020-01", 0.0)])),
                ]
            )

    def test_get_cost_and_usage_total(self):
        cost_explorer = CostExplorer()
        with patch.object(CostExplorerClient, "get_cost_and_usage") as mock_foo:
            mock_foo.return_value = [
                {
                    "TimePeriod": {"Start": "2019-12-01", "End": "2020-01-01"},
                    "Total": {
                        "UnblendedCost": {"Amount": "72.2197813571", "Unit": "USD"}
                    },
                    "Groups": [],
                    "Estimated": False,
                },
                {
                    "TimePeriod": {"Start": "2020-01-01", "End": "2020-02-01"},
                    "Total": {
                        "UnblendedCost": {"Amount": "68.0906860747", "Unit": "USD"}
                    },
                    "Groups": [],
                    "Estimated": False,
                },
                {
                    "TimePeriod": {"Start": "2020-02-01", "End": "2020-02-14"},
                    "Total": {
                        "UnblendedCost": {"Amount": "32.2073391037", "Unit": "USD"}
                    },
                    "Groups": [],
                    "Estimated": True,
                },
            ]
            print(cost_explorer.get_cost_and_usage_total())
            assert cost_explorer.get_cost_and_usage_total() == OrderedDict(
                [
                    (
                        "Total",
                        OrderedDict(
                            [("2019-12", 72.22), ("2020-01", 68.09), ("2020-02", 32.21)]
                        ),
                    )
                ]
            )
